from . import enums
from .enums import ui_frametype, ui_postype, ui_filltype
from pygame.locals import *
from . import models
from . import video
from OpenGL.GL import *
from . import inputs
from . import depgraph
from . import gamedata
import itertools
import inspect

class UITheme(object):
	def __init__(self):
		self.font = "micross20"


class UIFrame(object):
	def __init__(self, content, contentdata):
		self.func = None
		self.content = content
		self.data = contentdata
		self.parent = None
		self.minsize = None
		self.maxsize = None
		self.size = None
		self.gridsize = None
		self.pos = None
		self.postype = ui_postype.relative
		self.posset = None
		self.fill = ui_filltype.stretch
		self.draggable = False
		self.padding = (0, 0)
		self.color = None
		if content == ui_frametype.grid:
			self.set_stretch_equal()
		else:
			self.stretchiness = None

		
	def constrain_size(self, x, y):
		print("starting constrained to %s, %s" % (x, y))

		if self.content == ui_frametype.grid:
			x += self.padding[0] * len(self.data) * 2
			y += self.padding[1] * len(self.data[0]) * 2
			print("Added %s, %s of padding %s, %s" % (x, y, self.padding[0], self.padding[1]))
		else:
			x += self.padding[0] * 2
			y += self.padding[1] * 2
		
		if self.minsize:
			x = max(self.minsize[0], x)
			y = max(self.minsize[1], y)
		if self.maxsize:
			x = min(self.maxsize[0], x)
			y = min(self.maxsize[1], y)
		print("Constrained to %s, %s" % (x, y))
		return (x, y)
		
	
			
	def get_parent_size(self):
		if self.parent == None:
			return video.get_res()
		else:
			#return self.parent.get_child_size(self)
			return self.parent.size

	def get_child_size(self, child):
		if self.content == ui_frametype.grid:
			for i, row in enumerate(self.data):
				for j, col in enumerate(row):
					if col == child:
						x = self.gridsize[0][j]
						y = self.gridsize[1][i]
						# not sure why I originally disabled this but it seems to work ok for now?
						#if self.stretchiness:
						#	xlo = self.size[0] - sum(self.gridsize[0])
						#	ylo = self.size[1] - sum(self.gridsize[1])
						#	if xlo > 0:
						#		xs = self.stretchiness[0][j]
						#		xst = sum(self.stretchiness[0])
						#		x += int(float(xlo) * float(xs) / float(xst))
						#	if ylo > 0:
						#		ys = self.stretchiness[1][i]
						#		yst = sum(self.stretchiness[1])
						#		y += int(float(ylo) * float(ys) / float(yst))
						#
						return (x, y)
					
		else:
			return self.size
		

	def get_parent_pos(self):
		if self.parent == None:
			return (0, 0)
		else:
			#return self.parent.get_child_pos(self)
			return self.parent.pos
			
	def get_child_pos(self, child):
		if self.content == ui_frametype.grid:
			for i, row in enumerate(self.data):
				for j, col in enumerate(row):
					if col == child:
						#x = sum(self.gridsize[0][:j])+self.padding[0]
						#y = sum(self.gridsize[1][:i])+self.padding[1]
						x = sum(self.gridsize[0][:j])
						y = sum(self.gridsize[1][:i])
						return (x, y)
					
		else:
			return (0, 0)
			
		
	def set_stretch_cell(self, col, row):
		if self.content != ui_frametype.grid:
			raise TypeError("Stretch ratios only apply to grid type UIFrames")
		stretchiness_col = [0] * len(self.data)
		stretchiness_row = [0] * len(self.data[0])

		stretchiness_col[col] = 1
		stretchiness_row[row] = 1

		self.set_stretch_ratios(stretchiness_col, stretchiness_row)

	def set_stretch_equal(self):
		if self.content != ui_frametype.grid:
			raise TypeError("Stretch ratios only apply to grid type UIFrames")
		stretchiness_col = [1] * len(self.data)
		stretchiness_row = [1] * len(self.data[0])
		
		self.set_stretch_ratios(stretchiness_col, stretchiness_row)


	def set_stretch_ratios(self, stretchiness_col, stretchiness_row):
		if self.content != ui_frametype.grid:
			raise TypeError("Stretch ratios only apply to grid type UIFrames")
		if len(self.data) != len(stretchiness_col):
			raise ValueError("Grid has %d columns, but %d stretch ratios were provided" % (len(self.data), len(stretchiness_col)))
		if len(self.data[0]) != len(stretchiness_row):
			raise ValueError("Grid has %d rows, but %d stretch ratios were provided" % (len(self.data[0]), len(stretchiness_row)))

		stretchiness_col = [float(x) / float(sum(stretchiness_col)) for x in stretchiness_col]
		stretchiness_row = [float(y) / float(sum(stretchiness_row)) for y in stretchiness_row]
		
		self.stretchiness = (tuple(stretchiness_col), tuple(stretchiness_row))
		
	def autopos(self):
		if self.postype == ui_postype.center:
			ppos = self.get_parent_pos()
			psize = self.get_parent_size()
			
			self.pos = (ppos[0] + (psize[0]//2) - (self.size[0]//2), ppos[1] + (psize[1]//2) - (self.size[1]//2))
		elif self.postype == ui_postype.relative:
			if not self.posset:
				self.posset = self.parent.get_child_pos(self)
			ppos = self.get_parent_pos()
			psize = self.get_parent_size()
			if self.posset[0] >= 0:
				x = self.posset[0] + ppos[0]
			else:
				x = self.posset[0] - ppos[0] - psize[0] + self.size[0]
			if self.posset[1] >= 0:
				y = self.posset[1] + ppos[1]
			else:
				y = self.posset[1] - ppos[1] - psize[1] + self.size[1]
			self.pos = (x, y)
		elif self.postype == ui_postype.absolute:
			"This is bad. Don't do this. Asshole."
			self.pos = self.posset

		alist = [x for x in self.ancestors()]
		#if mgr.debugobj in alist:
		#	print(("FUCK. OK DUDER. I AM YOUR DEBUG OBJECT: %s" % (alist.index(mgr.debugobj),)))
		#	print((self.parent.pos))
		#	print((self.get_parent_pos()))
		#	print((self.parent.get_child_pos(self)))
		#	#print self.fill == ui_filltype.none
		#	#print self.postype == ui_postype.center
		#	print((self.pos))
		#	#print self.gridsize
		#	
		#	print("END DEBUG OBJECT")

			
		"""calculate position top-down, since child's position will depend on having an accurate parent position?"""
		for child in self:
			child.autopos()
		
			
	
	def autosize(self):
		if self.content == ui_frametype.model:
			self.size = self.constrain_size(int(self.data.get_size()[0]), int(self.data.get_size()[1]))
		elif self.content == ui_frametype.grid:
			if not self.data:
				self.size = self.constrain_size(0, 0)
				self.gridsize = [[], []]
				return
				
			self.gridsize = [[0] * len(self.data[0]), [0] * len(self.data)]
			for ir, row in enumerate(self.data):
				for ic, col in enumerate(row):
					col.autosize()
					self.gridsize[0][ic] = max(self.gridsize[0][ic], col.size[0])
					self.gridsize[1][ir] = max(self.gridsize[1][ir], col.size[1])
					print(("cell autosize: %s, %s" % (self.gridsize[1][ir], self.gridsize[0][ic])))
					
			self.size = self.constrain_size(sum(self.gridsize[0]), sum(self.gridsize[1]))
			print(("grid CONSTRAINED: %s - %s" % (self.size, self.padding)))
					
		elif self.content == ui_frametype.composite:
			x = 0
			y = 0
			for child in self:
				child.autosize()
				x = max(child.size[0], x)
				y = max(child.size[1], y)
				
			self.size = self.constrain_size(x, y)
			print(("composite CONSTRAINED: %s - %s" % (self.size, self.padding)))
				
	def autofillsize(self):
		""" autosize calculates absolute minimum sizes. """
		""" autofillsize then pads those as needed (particularly for grid cells) """

		if self.parent != None and self.fill != ui_filltype.none and self.postype != ui_postype.center:
			self.size = self.parent.get_child_size(self)
			
				
		if self.fill != ui_filltype.none and self.content == ui_frametype.grid: 
			print(("Hello I am grid %s, and my size is %s" % (self.gridsize, self.size)))
			if self.stretchiness:
				for ir, row in enumerate(self.data):
					for ic, col in enumerate(row):
						xlo = self.size[0] - sum(self.gridsize[0])
						ylo = self.size[1] - sum(self.gridsize[1])
						#print(("Stretching cell %s from %s to %s" % ((ic, ir), self.size, (self.size[0]+xlo, self.size[1]+ylo))))
						x = 0
						y = 0
						if xlo > 0:
							xs = self.stretchiness[0][ic]
							x = int(float(xlo) * xs)
						if ylo > 0:
							ys = self.stretchiness[1][ir]
							y = int(float(ylo) * ys)
							
						self.gridsize[0][ic] += x
						self.gridsize[1][ir] += y



		for child in self:
			child.autofillsize()
				
			
				
	def autoparent(self, parent=None):
		if parent != None:
			#print "Setting parent on %s to %s" % (self, parent)
			self.parent = parent
		for child in self:
			child.autoparent(self)
				
		
	def auto(self):
		self.autoparent()
		self.autosize()
		self.autofillsize()
		self.autopos()
	
	def ancestors(self):
		p = self.parent
		while True:
			if p == None:
				raise StopIteration()
			yield p
			p = p.parent
	
	def __iter__(self):
		if self.content == ui_frametype.model:
			return
			
		if self.content == ui_frametype.grid:
			for row in self.data:
				for col in row:
					yield col
		elif self.content == ui_frametype.composite:
			for child in self.data:
				yield child
				
	def __len__(self):
		if self.content == ui_frametype.model:
			return 0
			
		if self.content == ui_frametype.grid:
			if len(self.data) == 0:
				return 0
			return len(self.data) * len(self.data[0])
		elif self.content == ui_frametype.composite:
			return len(self.data)
			
	def handle_keypress(self, ev):
		return False
	
	def handle_event(self, ev):
		for child in self:
			handled = child.handle_event(ev)
			if handled:
				return True
				
		if self.relevant_event(ev):
			if ev.type == MOUSEBUTTONDOWN:
				self.set_focus()
				if self.draggable:
					self.drag_offset = (self.pos[0] - ev.pos[0], self.pos[1] - ev.pos[1])
					for p in self.ancestors():
						pass
						p
					mgr.drag_hook = p.move_drag
			elif ev.type == MOUSEBUTTONUP:
				if mgr.mousedown:
					#print "Mouseclick detected... %s %s ------ %s %s" % (mgr.mousedown, mgr.mousedown.pos if mgr.mousedown else None, self.contains_pos(mgr.mousedown.pos), self.func)
					if self.contains_pos(mgr.mousedown.pos):
						""" welp, this is a valid click -- down was inside us, and so is up """
						if self.func != None:
							if type(self.func) == str:
								mgr.get_callback(self.func)(ev)
							else:
								self.func(ev)
							return True
						
			elif ev.type == KEYUP:
				if self.is_focus():
					return self.handle_keypress(ev)
			elif ev.type == VIDEORESIZE:
				if self.parent == None:
					self.autosize()
					self.autopos()
			
			
				
				
		return False
		
	def convert_coords(self, pos, postype, newtype):
		if postype == newtype:
			return pos
		ppos = self.get_parent_pos()
		psize = self.get_parent_size()
		anchorx = False
		anchory = False
		if postype == ui_postype.relative:
			if pos[0] >= 0:
				x = pos[0] + ppos[0]
			else:
				x = pos[0] + ppos[0] + psize[0] - self.size[0]
			if pos[1] >= 0:
				y = pos[1] + ppos[1]
			else:
				y = pos[1] + ppos[1] + psize[1] - self.size[1]
			pos = (x, y)
		elif postype == ui_postype.center:
			pos = (ppos[0] + (psize[0]//2) - (self.size[0]//2), ppos[1] + (psize[1]//2) - (self.size[1]//2))

		if newtype == ui_postype.center:
			pos = (ppos[0] + (psize[0]//2) - (self.size[0]//2), ppos[1] + (psize[1]//2) - (self.size[1]//2))
		elif newtype == ui_postype.relative:
			if self.posset[0] >= 0:
				x = pos[0] - ppos[0]
			else:
				x = pos[0] - ppos[0] - psize[0] + self.size[0]
			if self.posset[1] >= 0:
				y = pos[1] - ppos[1]
			else:
				y = pos[1] - ppos[1] - psize[1] + self.size[1]
			
		elif newtype == ui_postype.absolute:
			pass
			
		return pos
		
	def move_drag(self, ev):
		newpos = (ev.pos[0] + self.drag_offset[0], ev.pos[1] + self.drag_offset[1])
		rel_newpos = self.convert_coords(pos, ui_postype.absolute, self.postype)
		self.posset = rel_newpos
		self.autopos()
				
	
	def contains_pos(self, pos):
		if pos[0] >= self.pos[0] and pos[0] < (self.pos[0] + self.size[0]):
			if pos[1] >= self.pos[1] and pos[1] < (self.pos[1] + self.size[1]):
				return True
				
		return False
		
	def relevant_event(self, ev):
		if ev.type == MOUSEMOTION:
			return self.contains_pos(ev.pos)
		elif ev.type == MOUSEBUTTONDOWN:
			return self.contains_pos(ev.pos)
		elif ev.type == MOUSEBUTTONUP:
			return self.contains_pos(ev.pos)
		elif ev.type == KEYDOWN:
			return self.is_focus()
		elif ev.type == KEYUP:
			return self.is_focus()
		else:
			return True
			
		
		
		
	def is_focus(self):
		return mgr.focus == self
		
	def set_focus(self):
		mgr.set_focus(self)
		
	def render(self):
		for child in self:
			child.render()
			
		if self.content == ui_frametype.model:
			if mgr.debug < 0:
				mgr.debug += 1
				if isinstance(self.data, models.TextModel):
					print(("%s - %s - TEXT %s" % (self.pos, self.size, self.data.text)))
				else:
					for l in self.data.layers:
						print(("%s - %s - %s" % (self.pos, self.size, l.material.id)))
			glPushMatrix()
			if isinstance(self.data, models.TextModel):
				glTranslate(self.pos[0], self.pos[1], 10.0)
			else:
				glTranslate(float(self.pos[0]) + (float(self.size[0]) / 2.0), float(self.pos[1]) + (float(self.size[1]) / 2.0), 10.0)

			newscale = [float(self.size[0]), float(self.size[1]), 1.0]
			for layer in self.data.layers:
				#if newscale != layer.scale:
				#	print "Changing scale on %s %s from %s to %s" % (layer.material.id, self.size, layer.scale, newscale)
				layer.scale = newscale
				#if "_bg_" in layer.material.id:
				#layer.color = self.color
			self.data.render()
			glPopMatrix()
		
		
#THIS IS OLD EXAMPLE CODE
#
#def startgame(ev):
#	mgr.remove("mainmenu")
#	
#def clicktest(ev):
#	print("I'm an event!")
#	
#def exitgame(ev):
#	inputs.add_command('exit')
#	
#def make_mainmenu():
#	b1 = make_button("Start", startgame)
#	b2 = make_button("Options", clicktest)
#	b3 = make_button("Exit", exitgame)
#
#	b1.padding = (20,0)
#	b2.padding = (20,0)
#	b3.padding = (20,0)
#	buttongrid = UIFrame(ui_frametype.grid, [[b1], [b2], [b3]])
#	buttongrid.postype = ui_postype.center
#	buttongrid.fill = ui_filltype.stretch
#	buttongrid.stretchiness = [[1], [1, 1, 1]]
#	mgr.debugobj = b1
#	buttongrid = UIFrame(ui_frametype.composite, [buttongrid])
#	buttongrid.padding = (20, 30)
#	mainframe = make_box("ui_frame", buttongrid, (0.0, 0.0, 1.0, 0.35))
#	mainframe.postype = ui_postype.center
#	return mainframe
#
#END OLD EXAMPLE CODE
#

class UIManager(object):
	def __init__(self):
		self.frames = []
		self.framenames = {}
		self.frame_library = {}
		self.focus = None
		self.last_mousedown = None
		self.mousemove_hook = None
		self.debug = 0
		self.theme = UITheme()
		self.callback_config_list = []
		self.callback_configs = {}

	def set_theme(self, theme):
		self.theme = theme
	
	def set_focus(self, frame):
		self.focus = frame
		
	def add(self, frame, name):
		assert name not in self.framenames
		self.framenames[name] = frame
		frame.index = len(self.frames)
		frame.auto()
		#print((frame.pos))
		#for child in frame:
		#	print(("%s-%s" % (child.pos, child.size)))
		"""	
		print "Ok"
		print frame.data[1][1].data[1].gridsize
		btngrid = frame.data[1][1].data[1]
		btn3 = btngrid.data[2][0]
		print "btn3"
		print btn3.pos
		print btngrid.get_child_pos(btn3)
		"""
		self.frames.append(frame)
		
	def remove(self, frame):
		fdel = self.framenames[frame]
		for i, f in enumerate(self.frames):
			if f is fdel:
				del self.frames[i]
		
	
		
	def render(self):
		for frame in self.frames:
			frame.render()

	def handle_event(self, ev):
		if ev.type == MOUSEMOTION and self.mousemove_hook != None:
			self.drag_hook(ev)
			return True
			
		handled = False
		for fr in self.frames:
			handled = fr.handle_event(ev)
			if handled:
				break
		if not handled:
			self.set_focus(None)

		if ev.type == MOUSEBUTTONDOWN:
			print(("MOUSECLICK: %s" % (ev.pos,)))
			self.mousedown = ev
		if ev.type == MOUSEBUTTONUP:
			self.drag_hook = None
			self.mousedown = None


	def make_box(self, sprite, center, color):
		l = [['tl', 'tc', 'tr'], ['cl', 'cc', 'cr'], ['bl', 'bc', 'br']]
		for i, r in enumerate(l):
			for j, c in enumerate(r):
				mdl = models.mgr.create("%s_%s" % (sprite, c))

				l[i][j] = UIFrame(ui_frametype.model, mdl)
				#if len(mdl.layers) > 1:
				mdl.layers[-1].color = color
				if c == 'cc':
					l[i][j] = UIFrame(ui_frametype.composite, [l[i][j], center])
				elif 'c' in c:
					"""Border edges should be repeated to fill!"""
					l[i][j].fill = ui_filltype.tile
					
					
			#l[i] = tuple(l[i])
		#l = tuple(l)
		uf = UIFrame(ui_frametype.grid, l)
		#uf.set_stretch_ratios((0, 1, 0), (0, 1, 0))
		uf.set_stretch_cell(1,1)
		return uf
		

	def make_button(self, text, clickfunc, padding=(0,0)):
		text = models.TextModel(self.theme.font, text)
		textframe = UIFrame(ui_frametype.model, text)
		textframe.postype = ui_postype.center
		textframe.filltype = ui_filltype.none

		paddingframe = UIFrame(ui_frametype.composite, [textframe])
		paddingframe.padding = padding
		paddingframe.postype = ui_postype.center

		btnframe = self.make_box("ui_button", paddingframe, (0.3, 0.3, 1.0, 1.0))
		btnframe.filltype = ui_filltype.none
		btnframe.func = clickfunc
		return btnframe


	def load_button(self, data):
		name = data["id"]
		print("loading button %s" % (name,))
		btn = self.make_button(data["label"], data["callback"], (data.padding["x"], data.padding["y"]))
		self.frame_library[name] = btn



	def load_frame(self, data):
		name = data["id"]
		print("loading frame %s" % (name,))
		framecolor = (data.color['r'], data.color['g'], data.color['b'], data.color['a'])
		frame = self.make_box("ui_frame", self.frame_library[data.content['id']], framecolor)
		frame.postype = enums.index(ui_postype, data['postype'])
		frame.filltype = enums.index(ui_filltype, data['filltype'])
		frame.padding = (data.padding['x'], data.padding['y'])

		self.frame_library[name] = frame

	def load_wrapper(self, data):
		name = data["id"]
		print("loading wrapper %s" % (name,))
		frame = UIFrame(ui_frametype.composite, [self.frame_library[data.content['id']]])
		frame.postype = enums.index(ui_postype, data['postype'])
		frame.filltype = enums.index(ui_filltype, data['filltype'])
		frame.padding = (data.padding['x'], data.padding['y'])
		
		self.frame_library[name] = frame

	def load_grid(self, data):
		name = data["id"]
		print("loading grid %s" % (name,))
		rowcols = []
		for r in data.row:
			rnum = r['number']
			rdata = []
			for c in r.col:
				cnum = c['number']
				rdata.append((cnum, self.frame_library[c.content['id']]))
			rdata = list(zip(*sorted(rdata)))[1]
			rowcols.append((rnum, rdata))
		rowcols = list(zip(*sorted(rowcols)))[1]

		frame = UIFrame(ui_frametype.grid, rowcols)
		frame.postype = enums.index(ui_postype, data['postype'])
		frame.filltype = enums.index(ui_filltype, data['filltype'])
		frame.padding = (data.padding['x'], data.padding['y'])

		self.frame_library[name] = frame

	def load_uidata(self):
		uidata = gamedata.get('ui')
		framelibdata = {}
		# start by preloading the names
		for data in itertools.chain(uidata.button, uidata.grid, uidata.frame, uidata.wrapper):
			framelibdata[data['id']] = data

		dg = depgraph.DepGraph()

		for data in uidata.button:
			framelibdata[data['id']] = (self.load_button, data)
			dg.add(data['id'], [])

		for data in uidata.wrapper:
			framelibdata[data['id']] = (self.load_wrapper, data)
			dg.add(data['id'], [data.content['id']])

		for data in uidata.frame:
			framelibdata[data['id']] = (self.load_frame, data)
			dg.add(data['id'], [data.content['id']])

		for data in uidata.grid:
			framelibdata[data['id']] = (self.load_grid, data)
			gcells = []
			for r in data.row:
				for c in r.col:
					gcells.append(c.content['id'])
			dg.add(data['id'], gcells)

		dg.resolve()

		for dn in dg.bottomup_iter():
			df, data = framelibdata[dn]
			df(data)


	def open_ui(self, name, callback_map, obj=None):
		self.callback_config_list.append(name)
		self.callback_configs[name] = (callback_map, obj)
		frame = self.frame_library[name]
		self.add(frame, name)

	def close_ui(self, name):
		i = self.callback_config_list.index(name)
		del self.callback_config_list[i]
		del self.callback_configs[name]
		self.remove(name)
		
		

	def get_callback(self, func):
		for config in self.callback_config_list:
			cc = self.callback_configs[config][0]
			if func in cc:
				return cc[func]

		raise KeyError("Callback function %s not defined" % (func,))

		

def init():
	global mgr
	mgr = UIManager()
	mgr.load_uidata()




mgr = None

